#!/usr/bin/env python3

from osgeo import gdal
import numpy as np
import os




###All values in degrees
spacing = 0.01
lat0 = 35.0
lon0 = 72.0
gridsize = 100

class Simulate:
    '''
    Object for each unit test.
    '''

    heading = { 'asc' : -12.0,
                'desc': -167.0}

    lookside = {'right' : -1,
                'left'  : 1}


    def __init__(self, passDirection, lookSide):
        '''
        Constructor with pass direction and lookside.
        '''

        self.passDirection = passDirection
        self.side = lookSide

    def simulateData(self):
        '''
        Simulate X and Y datasets.
        '''

        X = np.arange(-gridsize//2, gridsize//2 ) * spacing

        xx, yy = np.meshgrid(X,X)

        hdg = np.radians( self.heading[self.passDirection] )
        sgn = self.lookside[self.side]

        lat = lat0 + (yy * np.cos(hdg) -  xx * np.sin(hdg))
        lon = lon0 - sgn*(yy * np.sin(hdg) +  xx * np.cos(hdg))

        self.makeDatasets(lat, lon)


    def array2tif(self, arr, fname):
        '''
        Create geotiff from array.
        '''
        
        driver = gdal.GetDriverByName('GTiff')
        ds = driver.Create(fname, arr.shape[1], arr.shape[0], 1, gdal.GDT_Float64)
        bnd = ds.GetRasterBand(1)
        bnd.WriteArray(arr)
        bnd = None
        ds = None

    
    def makeDatasets(self, lat, lon):
        '''
        Create tiff files and vrt for the datasets.
        '''

        self.array2tif(lat, 'lat.tif')
        self.array2tif(lat, 'y.tif')
        self.array2tif(lon, 'lon.tif')
        self.array2tif(lon, 'x.tif')

        self.writeVRT( 'y.tif', 'lon.tif', 'lat.tif', 'yy.vrt')
        self.writeVRT( 'x.tif', 'lon.tif', 'lat.tif', 'xx.vrt')

        self.warp()

    def writeVRT(self, srcfile, lonfile, latfile, vrtfile):
        '''
        Write a VRT with src file and geoloc arrays.
        '''

        tmpl = '''
<VRTDataset RasterXSize="{size}" RasterYSize="{size}">
  <SRS>EPSG:4326</SRS>
  <VRTRasterBand band="1" dataType="Float64">
    <SimpleSource>
        <SourceFilename relativeToVRT="1">{src}</SourceFilename>
        <SourceBand>1</SourceBand>
        <SourceProperties RasterXSize="{size}" RasterYSize="{size}" DataType="Float64"/>
        <SrcRect xOff="0" yOff="0" xSize="{size}" ySize="{size}"/>
        <DstRect xOff="0" yOff="0" xSize="{size}" ySize="{size}"/>
    </SimpleSource>
  </VRTRasterBand>
  <metadata domain="GEOLOCATION">
    <mdi key="X_BAND">1</mdi>
    <mdi key="Y_BAND">1</mdi>
    <mdi key="LINE_STEP">1</mdi>
    <mdi key="PIXEL_STEP">1</mdi>
    <mdi key="X_DATASET">{lon}</mdi>
    <mdi key="PIXEL_OFFSET">0</mdi>
    <mdi key="Y_DATASET">{lat}</mdi>
    <mdi key="LINE_OFFSET">0</mdi>
  </metadata>
</VRTDataset>'''.format(size=gridsize, src=srcfile,
                        lon=lonfile, lat=latfile)

        with open(vrtfile, 'w') as fid:
            fid.write(tmpl)


    def cleanup(self):
        '''
        Clean up datasets.
        '''

        os.remove('lat.tif')
        os.remove('lon.tif')
        os.remove('x.tif')
        os.remove('xx.tif')
        os.remove('y.tif')
        os.remove('yy.tif')
        os.remove('xx.vrt')
        os.remove('yy.vrt')


    def warp(self):
        '''
        Use gdalwarp to warp the datasets.
        '''

        cmd = 'gdalwarp -r cubic -tr {2} {2} {0} {1}'

        status = os.system( cmd.format('xx.vrt', 'xx.tif', spacing/4))
        if status:
            raise Exception('X warp failed')


        status = os.system( cmd.format('yy.vrt', 'yy.tif', spacing/4))
        if status:
            raise Exception('Y warp failed')

    
    def analyze(self):
        '''
        Look at histogram of X and Y.
        '''

        ##Y dataset
        ds = gdal.Open('yy.tif', gdal.GA_ReadOnly)
        trans = ds.GetGeoTransform()
        data = ds.GetRasterBand(1).ReadAsArray()
        data[ data == 0 ] = np.nan

        ###Center pixel interpretation
        y = (trans[3] + 0.5 * trans[5]) + np.arange(ds.RasterYSize) * trans[5]
        x = (trans[0] + 0.5 * trans[1]) + np.arange(ds.RasterXSize) * trans[1]

        xx,yy = np.meshgrid(x,y)

        yy[ yy == 0 ] = np.nan
        residual = data - yy

        print('Y geolocation analysis:')
        print('Y Valid points : {0} '.format( np.sum( np.isfinite(residual))))
        print('Y spacing (deg): {0} '.format( trans[5] ))
        print('Y Mean    (pix): {0} '.format(np.nanmean(residual)/ trans[5]))
        print('Y Std dev (pix): {0} '.format(np.nanstd(residual) / np.abs(trans[5])))

        data = None
        ds = None
        xx = None
        yy = None


        ##X dataset
        ds = gdal.Open('xx.tif', gdal.GA_ReadOnly)
        trans = ds.GetGeoTransform()
        data = ds.GetRasterBand(1).ReadAsArray()
        data[ data == 0 ] = np.nan

        ###Center pixel interpretation
        y = (trans[3] + 0.5 * trans[5]) + np.arange(ds.RasterYSize) * trans[5]
        x = (trans[0] + 0.5 * trans[1]) + np.arange(ds.RasterXSize) * trans[1]

        xx,yy = np.meshgrid(x,y)

        xx[xx == 0] = np.nan
        residual = data - xx

        print('X geolocation analysis:')
        print('X Valid points : {0} '.format( np.sum( np.isfinite(residual))))
        print('X spacing (deg): {0} '.format( trans[1]))
        print('X Mean    (pix): {0} '.format(np.nanmean(residual)/ trans[1]))
        print('X Std dev (pix): {0} '.format(np.nanstd(residual)/trans[1]))

        data = None
        ds = None
        xx = None
        yy = None

        print('\n\n\n')


if __name__ == '__main__':
    '''
    Test driver.
    '''


    for passDir in Simulate.heading.keys():
        for side in Simulate.lookside.keys():
    
#    for passDir in ['desc',]:
#        for side in ['left',]:

            print('Pass Direction: {0} , Look Side: {1}'.format(passDir, side))
            sim = Simulate(passDir, side)
            sim.simulateData()
            sim.analyze()
            sim.cleanup()

