#!/usr/bin/env python3

from osgeo import gdal
import numpy as np
import os




###All values in degrees
spacing = 0.02
lat0 = 35.0
lon0 = 72.0
gridsize = 100

class Simulate:
    '''
    Object for each unit test.
    '''

    orientation = ['WENS', 'WESN', 'EWNS', 'EWSN']

    def __init__(self, orientation):
        '''
        Constructor with pass direction and lookside.
        '''

        if orientation.upper() in ['WENS', 'WESN']:
            self.xDirection = 1
        else:
            self.xDirection = -1

        if orientation.upper() in ['WENS', 'EWNS']:
            self.yDirection = -1
        else:
            self.yDirection = 1

        self.orientation = orientation.upper()

    def simulateData(self):
        '''
        Simulate X and Y datasets.
        '''

        X = lon0 + np.arange(gridsize) * spacing
        Y = lat0 + np.arange(gridsize) * spacing

        if self.xDirection == -1:
            X = X[::-1].copy()

        if self.yDirection == -1:
            Y = Y[::-1].copy()

        lon, lat = np.meshgrid(X,Y)

        self.makeDatasets(lat, lon)


    def array2tif(self, arr, fname):
        '''
        Create geotiff from array.
        '''
        
        driver = gdal.GetDriverByName('GTiff')
        ds = driver.Create(fname, arr.shape[1], arr.shape[0], 1, gdal.GDT_Float64)
        bnd = ds.GetRasterBand(1)
        bnd.WriteArray(arr)
        bnd = None
        ds = None

    
    def makeDatasets(self, lat, lon):
        '''
        Create tiff files and vrt for the datasets.
        '''
        os.makedirs(self.orientation)

        self.array2tif(lat, os.path.join(self.orientation, 'lat.tif'))
        self.array2tif(lat, os.path.join(self.orientation, 'y.tif'))
        self.array2tif(lon, os.path.join(self.orientation, 'lon.tif'))
        self.array2tif(lon, os.path.join(self.orientation, 'x.tif'))

        self.writeVRT( 'y.tif', 'lon.tif', 'lat.tif',
                       os.path.join(self.orientation, 'yy.vrt'))

        self.writeVRT( 'x.tif', 'lon.tif', 'lat.tif',
                       os.path.join(self.orientation, 'xx.vrt'))

        self.warp()

    def writeVRT(self, srcfile, lonfile, latfile, vrtfile):
        '''
        Write a VRT with src file and geoloc arrays.
        '''

        tmpl = '''
<VRTDataset RasterXSize="{size}" RasterYSize="{size}">
  <SRS>EPSG:4326</SRS>
  <VRTRasterBand band="1" dataType="Float64">
    <SimpleSource>
        <SourceFilename relativeToVRT="1">{src}</SourceFilename>
        <SourceBand>1</SourceBand>
        <SourceProperties RasterXSize="{size}" RasterYSize="{size}" DataType="Float64"/>
        <SrcRect xOff="0" yOff="0" xSize="{size}" ySize="{size}"/>
        <DstRect xOff="0" yOff="0" xSize="{size}" ySize="{size}"/>
    </SimpleSource>
  </VRTRasterBand>
  <metadata domain="GEOLOCATION">
    <mdi key="X_BAND">1</mdi>
    <mdi key="Y_BAND">1</mdi>
    <mdi key="LINE_STEP">1</mdi>
    <mdi key="PIXEL_STEP">1</mdi>
    <mdi key="X_DATASET">{lon}</mdi>
    <mdi key="PIXEL_OFFSET">0</mdi>
    <mdi key="Y_DATASET">{lat}</mdi>
    <mdi key="LINE_OFFSET">0</mdi>
  </metadata>
</VRTDataset>'''.format(size=gridsize, src=srcfile,
                        lon=lonfile, lat=latfile)

        with open(vrtfile, 'w') as fid:
            fid.write(tmpl)


    def cleanup(self):
        '''
        Clean up datasets.
        '''

        pwd = os.getcwd()
        os.chdir(self.orientation)
        os.remove('lat.tif')
        os.remove('lon.tif')
        os.remove('x.tif')
        os.remove('xx.tif')
        os.remove('y.tif')
        os.remove('yy.tif')
        os.remove('xx.vrt')
        os.remove('yy.vrt')
        os.remove( self.orientation + '.png')
        
        os.chdir(pwd)
        os.rmdir(self.orientation)

    def warp(self):
        '''
        Use gdalwarp to warp the datasets.
        '''
        pwd = os.getcwd()

        os.chdir(self.orientation)

#        cmd = '/Users/agram/packages/gdal/build/bin/gdalwarp -q -tap -r cubic -tr {2} {2} {0} {1}'
        cmd = 'gdalwarp -q -tap -r cubic -tr {2} {2} {0} {1}'
        status = os.system( cmd.format('xx.vrt', 'xx.tif', spacing))
        if status:
            raise Exception('X warp failed')


        status = os.system( cmd.format('yy.vrt', 'yy.tif', spacing))
        if status:
            raise Exception('Y warp failed')

        os.chdir(pwd)


    def analyze(self, plot=False):
        '''
        Look at histogram of X and Y.
        '''

        ##Y dataset
        ds = gdal.Open(os.path.join(self.orientation, 'yy.tif'), gdal.GA_ReadOnly)
        trans = ds.GetGeoTransform()
        data = ds.GetRasterBand(1).ReadAsArray()
        data[ data == 0 ] = np.nan

        ###Center pixel interpretation
        y = (trans[3] + 0.5 * trans[5]) + np.arange(ds.RasterYSize) * trans[5]
        x = (trans[0] + 0.5 * trans[1]) + np.arange(ds.RasterXSize) * trans[1]

        xx,yy = np.meshgrid(x,y)

        yy[ yy == 0 ] = np.nan
        yresidual = (data - yy)[2:-2,2:-2]

        print('Y geolocation analysis:')
        print('Y Valid points : {0} '.format( np.sum( np.isfinite(yresidual))))
        print('Y spacing (deg): {0} '.format( trans[5] ))
        print('Y median  (pix): {0} '.format( np.nanmedian(yresidual) / trans[5]))
        print('Y Mean    (pix): {0} '.format(np.nanmean(yresidual)/ trans[5]))
        print('Y Std dev (pix): {0} '.format(np.nanstd(yresidual) / np.abs(trans[5])))
        print('Y Origin : {0},{1} '.format(trans[0],trans[3]))

        yresidual /= np.abs(trans[5])
        data = None
        ds = None
        xx = None
        yy = None


        ##X dataset
        ds = gdal.Open(os.path.join(self.orientation, 'xx.tif'), gdal.GA_ReadOnly)
        trans = ds.GetGeoTransform()
        data = ds.GetRasterBand(1).ReadAsArray()
        data[ data == 0 ] = np.nan

        ###Center pixel interpretation
        y = (trans[3] + 0.5 * trans[5]) + np.arange(ds.RasterYSize) * trans[5]
        x = (trans[0] + 0.5 * trans[1]) + np.arange(ds.RasterXSize) * trans[1]

        xx,yy = np.meshgrid(x,y)

        xx[xx == 0] = np.nan
        xresidual = (data - xx)[2:-2,2:-2]

        print('X geolocation analysis:')
        print('X Valid points : {0} '.format( np.sum( np.isfinite(xresidual))))
        print('X spacing (deg): {0} '.format( trans[1]))
        print('X median (pix) : {0} '.format( np.nanmedian(xresidual.flatten())/trans[1]))
        print('X Mean    (pix): {0} '.format(np.nanmean(xresidual)/ trans[1]))
        print('X Std dev (pix): {0} '.format(np.nanstd(xresidual)/trans[1]))
        print('X Origin : {0},{1} '.format( trans[0], trans[3]))

        xresidual /= np.abs(trans[1])
        data = None
        ds = None
        xx = None
        yy = None

        print('\n\n\n')


        import matplotlib.pyplot as plt
        plt.figure(self.orientation)
        plt.subplot(1,2,1)
        plt.imshow(yresidual)
        plt.colorbar()

        plt.subplot(1,2,2)
        plt.imshow(xresidual)
        plt.colorbar()

        if plot:
            plt.show()
        else:
            plt.savefig( os.path.join(self.orientation, '{0}.png'.format(self.orientation)))


if __name__ == '__main__':
    '''
    Test driver.
    '''


    for passDir in Simulate.orientation:
    
        print('Orientation: {0}'.format(passDir))
        sim = Simulate(passDir)
        sim.simulateData()
        sim.analyze(plot=False)
#        sim.cleanup()

